// PatchWorks 1.0 Image ReadImage File
// This Software is released under the terms of the GPL.

// Reads a Windows Bitmap (*.BMP) (Truecolor) into a sImage structure.
// Maintainer: Steven Schummer
//
// return values: 1 = file could not be read
//				  2 = not a bitmap file	
//                3 = not a 24bit bitmap file
//                4 = not enough memory

#include "Image.h"



int ReadImage(sImage *srcImg, char *filename)
{
	FILE* fp;				// Filepointer
	ulong biSize=0;			// Header size - must be at least 40
	ulong biWidth=0;		// Image width in pixels
	ulong biHeight=0;		// Image height in pixels
	ushort biBM=0;			// Bitmap check
	ushort biPlanes=0;		// Must be 1
	ushort biBitCount=0;		// Bits per pixel - (1,2,4,8,16,24,32}
	ulong biCompression=0;		// Compression type (0 = uncompressed)
	ulong biSizeImage=0;		// Image size - uncompressed may be zero
	ulong biXPelsPerMeter=0;	// Preferred resolution (pixels/meter)
	ulong biYPelsPerMeter=0;	// Preferred resolution (pixels/meter)
	ulong sumByte=0;		// Size in bytes without headerinformation and fillbytes
	ulong fillBytes=0;		// Size in bytes of fillbytes in one row
	ulong biXdpi=0;			// Preferred resolution (pixels/inch)
	ulong biYdpi=0;			// Preferred resolution (pixels/inch)
	ulong i = 0;
	uint8 * readp;
	ulong count = 0;                // Counting variable for reading picture data
	ulong zCount = 0;		// Counting variable for reading picture data
	ulong pixRead = 0;		// Counting variable for reading picture data
	ulong bgrWidth = 0;		// Number of bytes in one row
	
	
	fp = fopen(filename, "rb"); // Opens a file (read binary)
	if(fp == NULL) return 1; // check if file could be read
	
	
	fseek(fp, 0, SEEK_SET);
	fread(&biBM, 2, 1, fp);
	
	if(biBM != 0x4D42) return 2;	//check for bitmap format, header entry
	
	fseek(fp, 28, SEEK_SET); // Go to the beginning of bitmapinfoheader
	fread(&biBitCount, 2, 1, fp); // Read bits per pixel value
	
	if(biBitCount != 24) return 3;
	
	fseek(fp, 14, SEEK_SET); // Go to beginning of info header
	fread(&biSize, 4, 1, fp); //Get headerinformation
	fread(&biWidth, 4, 1, fp);
	fread(&biHeight, 4, 1, fp);
	fread(&biPlanes, 2, 1, fp);
	fread(&biBitCount, 2, 1, fp);
	fread(&biCompression, 4, 1, fp);
	fread(&biSizeImage, 4, 1, fp);
	fread(&biXPelsPerMeter, 4, 1, fp);
	fread(&biYPelsPerMeter, 4, 1, fp);
	
	fseek(fp, 54, SEEK_SET); // Go to beginning of picture data
	
	// convert (pixel per meter) to (pixel per inch)
	biXdpi=(uint)(biXPelsPerMeter/(100/2.54f)+0.5); // +0.5 = rounding operation
	biYdpi=(uint)(biYPelsPerMeter/(100/2.54f)+0.5);
	
	// fill struct sImage with width, height, xdpi and ydpi
	srcImg->w=biWidth;
	srcImg->h=biHeight;
	srcImg->xdpi=biXdpi;
	srcImg->ydpi=biYdpi;
	
	sumByte = (srcImg->w * srcImg->h * 3); // Size in bytes without headerinformation and fillbytes
	fillBytes = (srcImg->w % 4); // Size in bytes of fillbytes in one row
	
	srcImg->data = malloc(sumByte); //allocate memory with sumByte bytes
	
	readp = srcImg->data;
	
	if(srcImg->data == NULL) return 4; //not enough memory
	
	bgrWidth = (srcImg->w) * 3; // number of bytes in one row
	
	// Read picture data in byte steps
	while(count < sumByte){
		while(zCount < bgrWidth) {
			pixRead = bgrWidth - zCount;
			zCount += fread(readp, 1, pixRead, fp);
			readp += zCount;
		}
		count += zCount;
		zCount = 0;
		fseek(fp, fillBytes, SEEK_CUR);
	}
	
	fclose(fp);
	
	return 0;
}

