// PatchWorks 1.0 Image WriteImage File
// This Software is released under the terms of the GPL.

// Writes an Image from an sImage structure into a Windows Bitmap file (*.BMP)
// Maintainer: Steven Schummer


#include "Image.h"
#include <malloc.h>	// malloc()
#include <string.h>	// strcopy()

typedef unsigned long ulong;
typedef unsigned short ushort;
typedef unsigned int uint;
typedef unsigned char uint8;

int WriteImage(sImage *srcImg, char *filename) {

	FILE* fp;
	ulong header4 = 0;	// 4 bytes for writing
	ushort header2 = 0;	// 2 bytes for writing
	uint8 leer = 0;		// Leerbyte
	ulong sumByte=0;	// Bytes without headerinformation and fillbytes
	ulong i = 0;
	ulong fillBytes=0;	// Bytes fillbytes
	uint8 * writep;		// picture data write pointer
	ulong count = 0;
	ulong zCount = 0;
	ulong bgrWidth = 0;
	ulong pixWrite = 0;
	ulong fileSize = 0;

	fillBytes = (srcImg->w % 4); // Get number of fillbytes per line
	sumByte = (srcImg->w * srcImg->h * 3); // Size in bytes of picture data
	bgrWidth = (srcImg->w) * 3; // Size of bytes of picture data per line

	writep = srcImg->data;
	
	fp = fopen(filename, "wb"); // Open file to write binary
	
	// Creating a bitmap fileformat
	// Description; Size 
	//
	// Signature; 2 bytes			 
	header2 = 0x4D42;
	fwrite(&header2,2,1,fp);

	// FileSize; 4 bytes
	fileSize = 54 + sumByte+fillBytes * srcImg->h;
	header4=fileSize+(fileSize%4);
	fwrite(&header4,4,1,fp);

	// Reserved; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// DataOffset; 4 bytes
	header4=54;
	fwrite(&header4,4,1,fp);

	// Size; 4 bytes
	header4=40;
	fwrite(&header4,4,1,fp);

	// Width; 4 bytes
	header4=srcImg->w;
	fwrite(&header4,4,1,fp);

	// Height; 4 bytes
	header4=srcImg->h;
	fwrite(&header4,4,1,fp);

	// Planes; 2 bytes
	header2=1;
	fwrite(&header2,2,1,fp);

	// BitCount; 2 bytes
	header2=24;
	fwrite(&header2,2,1,fp);

	// Compression; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// ImageSize; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	if( (srcImg->xdpi < PATCH_DPI_MINIMUM) || (srcImg->xdpi > PATCH_DPI_MAXIMUM) ) srcImg->xdpi = PATCH_DPI_DEFAULT;
	if( (srcImg->ydpi < PATCH_DPI_MINIMUM) || (srcImg->ydpi > PATCH_DPI_MAXIMUM) ) srcImg->ydpi = PATCH_DPI_DEFAULT;

	// XpixelsPerM; 4 bytes
	header4=(uint)((srcImg->xdpi*(100/2.54f))+0.5);
	fwrite(&header4,4,1,fp);

	// YpixelsPerM; 4 bytes
	header4=(uint)((srcImg->ydpi*(100/2.54f))+0.5);
	fwrite(&header4,4,1,fp);

	// ColorsUsed; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// ColorsImportant; 4 bytes
	header4=0;
	fwrite(&header4,4,1,fp);

	// picture data; 
	while(count < sumByte){
		while(zCount < bgrWidth) {
			pixWrite = bgrWidth - zCount;
			zCount += fwrite(writep, 1, pixWrite, fp);
			writep += zCount;
		}
		count += zCount;
		zCount = 0;
		for (i=0;i<fillBytes;i++) {
			fwrite(&leer, 1, 1, fp);
		}
	}
	// Add fillbytes at the end of picture data if necessary
	if(fileSize%4>0){
		for (i = 0; i < fileSize % 4; i++) {
			fwrite(&leer, 1, 1, fp);
		}
		
	}

	fclose(fp);
	return 0;
}