// PatchWorks 1.0 Patchwork CreatePatchwork File
// This Software is released under the terms of the GPL.

// Assembles a Patchwork, using the currently open Databases and the source image.
// Maintainer: Daniel Stiefelmaier


#include "Patchwork.h"

int FindFittingPatch(uint, uint, uint, sDataBase [], char **);

int CreatePatchwork(char *srcimg, char *dstimg, sDataBase DataBase[])
{
	sImage srcImg;
	sImage dstImg;
	sImage patImg; // temporary patch structure

	char *patchfile;
	
	uint r, g, b; // rgb-Values of source pixels
	uint i, j, k, l;
	uint err, dbinuse;
	uint srcoff, dstoff, dstoff_w, dstoff_h, dstoff_t, patoff;
	uint patch_w, patch_h;


	// read source image
	err = ReadImage(&srcImg, srcimg);
	if( err ) return 10 + err; // error while reading the source image

	// find a DB thats in use
	dbinuse = -1;
	for(i=0; i<MAX_OPEN_DBs; i++)
	{
		if(DataBase[i].inuse == 1)
		{
			dbinuse=i;
			break;
		}
	}
	if( dbinuse == -1 ) return 2;		// No DB in use


	// Patch resolution
	patch_w = DataBase[dbinuse].res_w;
	patch_h = DataBase[dbinuse].res_h;

	// Size of destination picture
	dstImg.w=srcImg.w*patch_w;
    dstImg.h=srcImg.h*patch_h;

	dstImg.data=malloc(dstImg.w*dstImg.h*3);	
	if(dstImg.data == NULL)	return 3;	// not enough memory


	// do conversion

	for(i=0; i<srcImg.h; i++) // vertical patch count
	{
		for(j=0; j<srcImg.w; j++) // hozizontal patch count
		{
			srcoff = (srcImg.w*i + j) * 3; //full lines plus current offset times 3 bytes

			b = srcImg.data[srcoff  ];
			g = srcImg.data[srcoff+1];
			r = srcImg.data[srcoff+2];

			err = FindFittingPatch(r, g, b, DataBase, &patchfile);
			if( err ) return 20 + err;
//printf("Image: %s\n", patchfile);
			err = ReadImage(&patImg, patchfile);
			if( err ) return 30 + err; // error while reading patch image
			free(patchfile);

			dstoff_h = i*srcImg.w*patch_w*patch_h*3;// total offset of fully-written patchlines
			dstoff_w = patch_w * j * 3;
			dstoff_t = dstoff_h + dstoff_w; // offset of FIRST byte of current patch
//printf("off: %i\n", dstoff_t);
			for(k=0; k<patch_h; k++)
			{
				dstoff = dstoff_t + k*dstImg.w*3; // add linewith * writtenlines
				patoff   = k*patch_w*3;

				for(l=0; l<patch_w*3; l++)
				{
					dstImg.data[dstoff+l] = patImg.data[patoff + l];
				}
			}

			free(patImg.data);
		}
	}

	// write source image
	err = WriteImage(&dstImg, dstimg);
	if( err ) return 10 + err; // error while reading the source image


	free(srcImg.data);
	free(dstImg.data);

	return 0;
}



typedef struct
{
	uint dbnumber;
	uint imgnumber;
	double assessment;

}sFitting;


// returns the path of the best fitting patch to a given pattern.
int FindFittingPatch(uint r, uint g, uint b, sDataBase DataBase[], char **fittingpath)
{
	uint k, i, o, err;
	int rtmp, gtmp, btmp; // temp colors rgb
	uint fitting;

	char filename[INDEX_FILE_LINE]; // long enough...
	int filenameoff;
	
	sFitting fittingcandidates[MAX_OPEN_DBs];
	uint candidatecount=0;	// counts fittingcandidates (= count of open DBs)

	double *assessment = NULL; // assessment, how propper a coler fits the given pattern. (int would be faster)


	for(k=0; k<MAX_OPEN_DBs; k++)
	{
		if(DataBase[k].inuse == 0) continue;  // try next DB
//printf("??:\n%s\n", DataBase[k].data);
		assessment = realloc(assessment, INDEX_FILE_LINE * DataBase[k].imgcount);
		if( assessment == NULL ) return 1; // not enough memory
		
		for(i=0; i<DataBase[k].imgcount; i++)
		{
			o = i*INDEX_FILE_LINE + 13; // offset. 13 = first byte of RRR, GGG, BBB

			err = sscanf(&DataBase[k].data[o], "%i %i %i", &rtmp, &gtmp, &btmp);
			if(err != 3) return 2;			// could not read all channels
			
			rtmp -= r;   // difference in r channel
			gtmp -= g;
			btmp -= b;
			
			if( rtmp < 0 ) rtmp *= (-1); // get value (Betrag)
			if( gtmp < 0 ) gtmp *= (-1);
			if( btmp < 0 ) btmp *= (-1);
			
			//assessment[i] = pow(2, rtmp) + pow(2, gtmp) + pow(2, btmp);
			assessment[i] = rtmp + gtmp + btmp;
		}
		
		
		// now get best fitting color
		fitting = 0; // set to first element at the beginning
		
		for(i=1; i<DataBase[k].imgcount; i++)
		{
			if(	assessment[i] < assessment[fitting] ) fitting = i;
		}

		// add to candidates
		fittingcandidates[candidatecount].dbnumber   = k;
		fittingcandidates[candidatecount].imgnumber  = fitting;
		fittingcandidates[candidatecount].assessment = assessment[i];
		candidatecount++;

	} // for all DBs


	// finally choose REALLY best fitting
	fitting = 0; // set to first element at the beginning

	for(i=0; i<candidatecount; i++)
	{
		if( fittingcandidates[i].assessment < fittingcandidates[i].assessment ) fitting = i;
	}

	i			= fittingcandidates[fitting].dbnumber;
	filenameoff	= fittingcandidates[fitting].imgnumber * INDEX_FILE_LINE;

	*fittingpath = malloc( strlen(DataBase[i].path) + INDEX_FILE_LINE );
	if( *fittingpath == NULL ) return 1; // not enough memory

	for(k=0; k<12; k++)	filename[k] = DataBase[i].data[filenameoff+k];
	filename[12]='\0';

	strcpy(*fittingpath, DataBase[i].path);

	PathMerger(*fittingpath, filename);

	free(assessment);
//	free(fittingpath); // not to be freed...

	return 0;
}