// PatchWorks 1.0 Patchwork CreatePseudoPatchwork File
// This Software is released under the terms of the GPL.

// Assembles a pseudo Patchwork using a specified number of random colors.
// Maintainer: Philippe Trolliet (and Daniel Stiefelmaier)

#include "Patchwork.h"

//by Daniel Stiefelmaier
int FindFittingPseudoPatch(uint, uint, uint, uint8 *, uint, uint *);

//by  Philippe Trolliet
int CreatePseudoPatchwork(char* srcImagePath, char* dstImagePath, uint patchResWidth, uint patchResHeight, uint colorCount) {
	sImage srcImg; // Struktur des Quellbilds
	sImage dstImg; // Struktur des Zielbilds
	uint fitting; // beinhaltet die am bestern passenden rgb-werte
	uint red; // r-Wert des Quell Pixels
	uint green; // g-Wert des Quell Pixels
	uint blue;						// b-Wert des Quell Pixels
	uint8* colors;					// colorCount Zufallsfarbwerte
	uint i=0, j=0, k=0, l=0, m=0;	// Laufvariablen fr for-Schleifen
	uint abortConditionSrc=0;		// dynamische Abbruchbedingung fr for-Schleife (Quellbild)
	uint abortConditionLine=0;		// dynamische Abbruchbedingung fr for-Schleife (Ziel Zeile)
	uint abortConditionDst=0;		// dynamische Abbruchbedingung fr for-Schleife (Zielbild)
	uint8* line;					// beinhaltet eine fertige Zeile des Zielbilds

	// Quellbild einlesen
	ReadImage(&srcImg, srcImagePath);


	// Grsse des Zielbilds berechnen und Speicherreservierung
	dstImg.w=srcImg.w*patchResWidth;
    dstImg.h=srcImg.h*patchResHeight;
	dstImg.data=malloc(dstImg.w*dstImg.h*3);
	// berprfen, ob Speicher reserviert werden konnte
	if(dstImg.data == NULL)
		return -1;

	colors=malloc(colorCount*3);
	if(colors == NULL)
		return -1;

	line=malloc(srcImg.w*patchResWidth*3);
	// berprfen, ob Speicher reserviert werden konnte
	if(line == NULL)
		return -1;


	srand((unsigned)time(NULL));

	// fllen von colors mit Zufallsfarbwerten
	for(i=0; i<colorCount*3; i++) {
		colors[i]=rand();
	}

	i=0;
	// neues zielbild konstruieren
	for(m=0; m<srcImg.h; m++) {
		abortConditionLine = 0;
		j=0;

		abortConditionSrc+=srcImg.w*3;
		// eine zeile des quellbilds in eine zeile vom zieldbild umwandeln

		for(i; i<abortConditionSrc; i+=3) {
			blue=srcImg.data[i];
			green=srcImg.data[i+1];
			red=srcImg.data[i+2];

			FindFittingPseudoPatch(red, green, blue, colors, colorCount, &fitting);

			abortConditionLine+=patchResWidth*3;

			// Quellpixel resPatchWidth mal mit fitting schreiben
			for(j; j<abortConditionLine; j+=3) {
				line[j]=colors[fitting*3];
				line[j+1]=colors[fitting*3+1];
				line[j+2]=colors[fitting*3+2];
			}
		}

		abortConditionDst+=dstImg.w*3*patchResHeight;

		// Zielzeile (line[]) resPatchHeight mal schreiben
		for(k; k<abortConditionDst; k+=3) {
			dstImg.data[k]=line[l];
			dstImg.data[k+1]=line[l+1];
			dstImg.data[k+2]=line[l+2];
			l+=3;
			// wenn Zeilenende erreicht, l=0 setzen
			//if(k % (dstImg.w*3)==0)	{l=0;}
	//bug correction:
			if(l % (dstImg.w*3)==0) l=0;

		}
	}
	WriteImage(&dstImg, dstImagePath);

	free(srcImg.data);
	free(dstImg.data);
	free(line);
	free(colors);

	return 0;
}




// returns best fitting color of a list of colors to a given pattern.
int FindFittingPseudoPatch(uint r, uint g, uint b, uint8 *colors, uint colorcount, uint *fitting)
{
	uint i, o, fits;
	int rtmp, gtmp, btmp; // temp colors rgb

	//double *assessment; // assessment, how propper a coler fits the given pattern.

	//assessment = malloc(sizeof(double)*colorcount);

	int *assessment;
	assessment = malloc(sizeof(int)*colorcount);

	if( assessment == NULL ) return 1; // not enough memory

	fits = 0; // set to first element at the beginning

	for(i=0; i<colorcount; i++)
	{
		o = i*3; // offset

		btmp = b - colors[o  ];   // difference in r channel
		gtmp = g - colors[o+1];
		rtmp = r - colors[o+2];

		if( btmp < 0 ) btmp *= (-1); // get absolute value (Betrag)
		if( gtmp < 0 ) gtmp *= (-1);
		if( rtmp < 0 ) rtmp *= (-1);

		//assessment[i] = pow(2, rtmp) + pow(2, gtmp) + pow(2, btmp);
		assessment[i] = rtmp + gtmp + btmp;
		//printf("r %i  g %i  b %i  ass %g\n", rtmp, gtmp, btmp, assessment[i]);

		// does current color fit better?
		if(	assessment[i] < assessment[fits] ) fits = i;
	}
	
	*fitting = fits; // save time by not alwasy use the dereferenced variable

	free(assessment);

	return 0;
}




/*

  // Stiefels Variante (nicht funktionstchtig, momentan)
int CreatePseudoPatchwork(char* srcImagePath, char* dstImagePath, uint patchResWidth, uint patchResHeight, uint colorCount)
{
	sImage srcImg; // Struktur des Quellbilds
	sImage dstImg; // Struktur des Zielbilds
	uint fitting; // beinhaltet die am bestern passenden rgb-werte
	uint red; // r-Wert des Quell Pixels
	uint green; // g-Wert des Quell Pixels
	uint blue; // b-Wert des Quell Pixels
	uint8* colors; // colorCount Zufallsfarbwerte
	uint i=0, j=0, k=0, l=0, m=0, n; // Laufvariablen fr for-Schleifen
	uint abortConditionSrc=0; // dynamische Abbruchbedingung fr for-Schleife (Quellbild)
	uint abortConditionLine=0; // dynamische Abbruchbedingung fr for-Schleife (Ziel Zeile)
	uint abortConditionDst=0; // dynamische Abbruchbedingung fr for-Schleife (Zielbild)
	uint8* line; // beinhaltet eine fertige Zeile des Zielbilds
	uint srcoff, srcvoff, dstoff, dstvoff;

	// Quellbild einlesen
	ReadImage(&srcImg, srcImagePath);

	// Grsse des Zielbilds berechnen und Speicherreservierung
	dstImg.w=srcImg.w*patchResWidth;
    dstImg.h=srcImg.h*patchResHeight;

	dstImg.data=malloc(dstImg.w*dstImg.h*3);	
	if(dstImg.data == NULL)	return 1;	// not enough memory

	colors=malloc(colorCount*3);
	if(colors == NULL)		return 1;	// not enough memory

	line=malloc(srcImg.w*patchResWidth*3);
	if(line == NULL)		return 1;	// not enough memory

	srand((unsigned)time(NULL));

	// befllen von colors mit Zufallsfarbwerten
	for(i=0; i<colorCount*3; i++) {
		colors[i]=rand();
	}

	colors[0] = 255;
	colors[1] = 0;
	colors[2] = 0;

	colors[3] = 0;
	colors[4] = 255;
	colors[5] = 0;

	colors[6] = 0;
	colors[7] = 0;
	colors[8] = 255;
	
	colors[9]  = 255;
	colors[10] = 255;
	colors[11] = 0;

	colors[12] = 0;
	colors[13] = 0;
	colors[14] = 0;

	// test
	//dstImg.w=srcImg.w;
    //dstImg.h=srcImg.h;
	//dstImg.data=malloc(dstImg.w*dstImg.h*3);
	//for(i=0; i<dstImg.w*dstImg.h*3; i++) dstImg.data[i] = srcImg.data[i];
	

/*
	// neues zielbild konstruieren
	for(m=0; m<srcImg.h; m++)
	{
		// eine zeile des quellbilds in eine zeile vom zieldbild umwandeln
		for(i=0; i< srcImg.w; i++)
		{
			blue	= srcImg.data[m*srcImg.w+i  ];
			green	= srcImg.data[m*srcImg.w+i+1];
			red		= srcImg.data[m*srcImg.w+i+2];

			FindFittingPseudoPatch(red, green, blue, colors, colorCount, &fitting);

			// Quellpixel resPatchWidth mal mit fitting schreiben
			for(j=0; j < patchResWidth; j++)
			{
				line[i*3*patchResWidth + j*3  ] = colors[fitting*3  ];
				line[i*3*patchResWidth + j*3+1] = colors[fitting*3+1];
				line[i*3*patchResWidth + j*3+2] = colors[fitting*3+2];
			}
		}	
		// Zielzeile (line[]) resPatchHeight mal schreiben
		for(k=0; k< patchResHeight; k++)
		{
			for(j=0; j< dstImg.w*3; j++)
			{
				dstImg.data[l++] = line[j];
			}
		}
	}
* /

	for(m=0; m<srcImg.h; m++)
	{
		srcvoff = m*srcImg.w*3; // lines*linesize

		for(n=0; n<srcImg.w; m++)
		{
			srcoff = srcvoff + n*3;

			blue	= srcImg.data[srcoff  ];
			green	= srcImg.data[srcoff+1];
			red		= srcImg.data[srcoff+2];
			
			FindFittingPseudoPatch(red, green, blue, colors, colorCount, &fitting);
			fitting*=3; // offset of fitting color
			
			dstvoff = dstImg.w*3 * patchResHeight*m; //linesize * writtenlines

			// write patch
			for(i=0; i < patchResHeight; i++)
			{
				dstoff = dstvoff + n*3*patchResWidth + i*dstImg.w; // fullywrittenpatches + patchesincurrentline + patchlineoffset
				
				for(j=0; j < patchResWidth; j++)
				{
					dstImg.data[dstoff + j*3  ] = colors[fitting  ];
					dstImg.data[dstoff + j*3+1] = colors[fitting+1];
					dstImg.data[dstoff + j*3+2] = colors[fitting+2];
				}
			}

		}
	}
		


	

	WriteImage(&dstImg, dstImagePath);

	free(srcImg.data);
	free(dstImg.data);
	free(line);
	free(colors);

	return 0;
}

*/
